#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <errno.h>

#include <sys/io.h>
#include <sys/mman.h>

volatile unsigned char stopit = 0;
void signalled(int sig)
{
   stopit = 1;
}

#define CACHE_SIZE   512*1024
#define CACHE_LINE_SIZE    64

void run_unshielded_half( void )
{
   unsigned char block[2*CACHE_SIZE];
   unsigned int i = 0;

   while (!stopit)
   {
      for(i=0; i<sizeof(block); i+= CACHE_LINE_SIZE)
         block[i] = (unsigned char)i;
   }
}

typedef unsigned long long u64;
typedef unsigned long u32;

inline u64 rdtsc() {
   u32 TSC_low, TSC_high;
   u64 TSC;
   __asm__ __volatile__ ("rdtsc" : "=a" (TSC_low), "=d" (TSC_high));
   TSC = TSC_high;
   TSC <<= 32;
   TSC += TSC_low;
   return TSC;
}

unsigned int histogram[1000];

void hist( void )
{
   int i;

   for(i=0; i<1000; i++)
   {
      if (histogram[i])
      {
         printf("%d %d\n",i,histogram[i]);
      }
   }
}

void run_shielded_half( void )
{
   unsigned char memloc = 0;

   while(!stopit)
   {
      __asm__ __volatile__ ("cli");
      u64 read1 = rdtsc();
      memloc++;
      u64 read2 = rdtsc();
      __asm__ __volatile__ ("sti");

      u64 delta = read2-read1;
      if (delta >= 1000) delta = 999;

      histogram[delta]++;
   }

   hist();
}

void usage( void )
{
   fprintf(stderr,"Usage: cache_test [1|2]\n");
   exit(1);
}

int main( int argc, char **argv )
{
   if (argc != 2)
   {
      usage();
   }

   if (iopl(3) != 0)
   {
      int error_id = errno;
      fprintf(stderr,"I/O priv. escalation failed: %s\n",strerror(error_id));
      exit(1);
   }

   if (mlockall(MCL_CURRENT|MCL_FUTURE) != 0) {
      perror("mlockall");
      exit(1);
   }

   signal(SIGINT, signalled);

   if (argv[1][0] == '1')
   {
      run_unshielded_half();
   }
   else if (argv[1][0] == '2')
   {
      run_shielded_half();
   }
   else
   {
      usage();
   }

   return 0;
}
