#include <linux/rtc.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sched.h>
#include <sys/signal.h>
#include <string.h>
#include <sys/io.h>

#define UNUSED(x)    ((void)x)

typedef unsigned long long u64;
typedef unsigned long u32;

typedef struct {
   unsigned int foo:!(sizeof(u64) - 8);
} test_struct;

inline u32 rdtscl() {
   u32 TSC;
   __asm__ __volatile__ ("rdtsc" : "=a" (TSC)::"edx");
   return TSC;
}

//TODO:jst: why is this not working?  Only returns low 32 bits
inline u64 rdtscll() {
   u64 TSC;
   __asm__ __volatile__ ("rdtsc" : "=A" (TSC));
   return TSC;
}

inline u64 rdtsc() {
   u32 TSC_low, TSC_high;
   u64 TSC;
   __asm__ __volatile__ ("rdtsc" : "=a" (TSC_low), "=d" (TSC_high));
   TSC = TSC_high;
   TSC <<= 32;
   TSC += TSC_low;
   return TSC;
}

inline void put_parallel_data( void )
{
   static unsigned char data = 0;

   outb(data++,0x378);
}

double second() {
   struct timeval tv;
   gettimeofday(&tv,0);
   return tv.tv_sec + 1e-6 * tv.tv_usec;
}

void selectsleep(unsigned us) {
   struct timeval tv;
   tv.tv_sec = 0;
   tv.tv_usec = us;
   select(0,0,0,0,&tv);
}

double secondsPerTick, ticksPerSecond;

void calibrate() {
   double sumx = 0;
   double sumy = 0;
   double sumxx = 0;
   double sumxy = 0;
   double slope;

   // least squares linear regression of ticks onto real time
   // as returned by gettimeofday.

   const unsigned n = 30;
   unsigned i;

   for (i=0; i<n; i++) {
      double breal,real,ticks;
      u64 bticks;

      breal = second();
      bticks = rdtsc();

      selectsleep((unsigned)(10000 + drand48() * 200000));

      ticks = rdtsc() - bticks;
      real = second() - breal;

      sumx += real;
      sumxx += real * real;
      sumxy += real * ticks;
      sumy += ticks;
   }
   slope = ((sumxy - (sumx*sumy) / n) /
       (sumxx - (sumx*sumx) / n));
   ticksPerSecond = slope;
   secondsPerTick = 1.0 / slope;
   printf("%3.3f MHz\n",ticksPerSecond*1e-6);
}

void fatal(char *msg) {
   perror(msg);
   exit(1);
}

int set_realtime_priority(void)
{
   struct sched_param schp;
   /*
    * set the process to realtime privs
    */
   memset(&schp, 0, sizeof(schp));
   schp.sched_priority = sched_get_priority_max(SCHED_FIFO);

   if (sched_setscheduler(0, SCHED_FIFO, &schp) != 0) {
      perror("sched_setscheduler");
      exit(1);
   }

   return 0;
}

volatile int stopit;

#define SAMPLES   100000
int histogram[SAMPLES];

void hist(char *file) {
   int i;
   u64 hits = 0;
   double denom;
   FILE *f;

   f = fopen(file, "w");
   if (f == 0) {
      fprintf(stderr, "realfeel: can't open `%s':%s\n",
         file, strerror(errno));
      exit(1);
   }

   for (i = 0; i < SAMPLES; i++)
   {
      hits += histogram[i];
   }

   denom = hits / 100.0;

   for (i = 0; i < SAMPLES; i++) {
      if (histogram[i]) {
         fprintf(f, "%d.%d, %9.6f\n",
            i/10,i%10, (double)(histogram[i] / denom));
      }
   }
   fclose(f);
   exit(0);
}

void signalled(int sig) {
   stopit = 1;
}

static void usage(void) {
   fprintf(stderr, "Usage: realfeel filename.hist\n");
   exit(1);
}

const unsigned char interrupts = 0;

int main( int argc, char ** argv )
{
   UNUSED(argc);
   UNUSED(argv);

   if (argc != 2)
      usage();

   if (mlockall(MCL_CURRENT|MCL_FUTURE) != 0) {
      perror("mlockall");
      exit(1);
   }
   set_realtime_priority();
   calibrate();
   printf("secondsPerTick=%f\n", secondsPerTick);
   printf("ticksPerSecond=%f\n", ticksPerSecond);

   if (iopl(3) != 0)
   {
      int error_id = errno;
      fprintf(stderr,"I/O priv. escalation failed: %s\n",strerror(error_id));
      return 1;
   }

   signal(SIGINT, signalled);

   u64 total = 0;
   u64 min = ~0;
   u64 max = 0;
   unsigned char reset = 0;

   while(!stopit)
   {
      if (!interrupts) __asm__ __volatile__ ("cli");
      u64 read1 = rdtsc();
      put_parallel_data();
      u64 read2 = rdtsc();
      if (!interrupts) __asm__ __volatile__ ("sti");    //user space...this is always safe

      u64 delta = read2-read1;

      double s = (delta / ticksPerSecond);
      double usecs_round = (unsigned int)(10000000.0*s) + 0.5;    //record tenths of usecs
      //printf("delta = %16ld (%.6fs == %dus)\n",delta,s,(unsigned int)usecs_round);

      if (usecs_round >= SAMPLES-1)
      {
         histogram[SAMPLES-1]++;
      }
      else
      {
         histogram[(unsigned int)usecs_round]++;
      }
   }

   hist(argv[1]);
   return 0;
}
